# future
from __future__ import annotations

# stdlib
import functools
from functools import lru_cache
import operator
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
import numpy as np
import torch

# syft absolute
# absolute
import syft as sy

# relative
from . import utils
from .... import logger
from ....grid import GridURL
from ....proto.core.tensor.share_tensor_pb2 import ShareTensor as ShareTensor_PB
from ...common.serde.deserialize import _deserialize as deserialize
from ...common.serde.serializable import serializable
from ...common.serde.serialize import _serialize as serialize
from ...smpc.store.crypto_store import CryptoStore
from ..passthrough import PassthroughTensor  # type: ignore

METHODS_FORWARD_ALL_SHARES = {
    "repeat",
    "copy",
    "diagonal",
    "flatten",
    "transpose",
    "partition",
    "resize",
    "ravel",
    "compress",
    "reshape",
    "squeeze",
    "swapaxes",
    "__pos__",
    "__neg__",
    "take",
    "choose",
    "cumsum",
    "trace",
}
INPLACE_OPS = {"resize", "put"}
RING_SIZE_TO_OP = {
    2: {
        "add": operator.xor,
        "sub": operator.xor,
        "mul": operator.and_,
        "lt": operator.lt,
        "gt": operator.gt,
        "ge": operator.ge,
        "le": operator.le,
        "eq": operator.eq,
        "ne": operator.ne,
    },
    2
    ** 32: {
        "add": operator.add,
        "sub": operator.sub,
        "mul": operator.mul,
        "lt": operator.lt,
        "gt": operator.gt,
        "ge": operator.ge,
        "le": operator.le,
        "eq": operator.eq,
        "ne": operator.ne,
    },
}

CACHE_CLIENTS: Dict[str, Any] = {}


def populate_store(*args: List[Any], **kwargs: Dict[Any, Any]) -> None:
    ShareTensor.crypto_store.populate_store(*args, **kwargs)  # type: ignore


@serializable()
class ShareTensor(PassthroughTensor):
    crypto_store = CryptoStore()

    __slots__ = (
        "rank",
        "ring_size",
        "clients",  # clients connections
        "min_value",
        "max_value",
        "generator_przs",
        # Only ShareTensors with seed_przs could be sent over the wire
        "seed_przs",
        "parties_info",
        "nr_parties",
    )

    def __init__(
        self,
        rank: int,
        parties_info: List[GridURL],
        ring_size: int,
        seed_przs: int = 42,
        clients: Optional[List[Any]] = None,
        value: Optional[Any] = None,
        init_clients: bool = False,
    ) -> None:
        # TODO: Ring size needs to be changed to 2^64 (or other specific sizes)
        self.rank = rank
        self.ring_size = ring_size
        self.nr_parties = len(parties_info)
        self.parties_info = parties_info
        self.clients = []
        if clients is not None:
            self.clients = clients
        elif init_clients:  # type: ignore
            self.clients = ShareTensor.login_clients(parties_info)

        self.min_value, self.max_value = ShareTensor.compute_min_max_from_ring(
            self.ring_size
        )

        # This should be set only in the deserializer
        self.generator_przs = None
        self.seed_przs = seed_przs
        super().__init__(value)

    @staticmethod
    def login_clients(parties_info: List[GridURL]) -> Any:
        clients = []
        for party_info in parties_info:
            # if its localhost make it docker-host otherwise no change
            external_host_info = party_info.as_docker_host()
            client = CACHE_CLIENTS.get(str(external_host_info), None)

            if client is None:
                # default cache to true, here to prevent multiple logins
                # due to gevent monkey patching, context switch is done during
                # during socket connection initialization.
                CACHE_CLIENTS[str(external_host_info)] = True
                # TODO: refactor to use a guest account
                client = sy.login(  # nosec
                    url=external_host_info,
                    email="info@openmined.org",
                    password="changethis",
                    port=external_host_info.port,
                    verbose=False,
                )
                CACHE_CLIENTS[str(external_host_info)] = client
            clients.append(client)
        return clients

    def __getitem__(self, item: Union[str, int, slice]) -> ShareTensor:
        return ShareTensor(
            rank=self.rank,
            parties_info=self.parties_info,
            ring_size=self.ring_size,
            value=self.child[item],
            clients=self.clients,
        )

    def copy_tensor(self) -> ShareTensor:
        return ShareTensor(
            value=self.child,
            rank=self.rank,
            parties_info=self.parties_info,
            ring_size=self.ring_size,
            seed_przs=self.seed_przs,
            clients=self.clients,
        )

    @staticmethod
    @lru_cache(32)
    def compute_min_max_from_ring(ring_size: int = 2**32) -> Tuple[int, int]:
        if ring_size == 2:
            min_value, max_value = 0, 1
        else:
            min_value = (-ring_size) // 2
            max_value = (ring_size) // 2 - 1
        return min_value, max_value

    @staticmethod
    @lru_cache(maxsize=None)
    def get_op(ring_size: int, op_str: str) -> Callable[..., Any]:
        """Returns method attribute based on ring_size and op_str.
        Args:
            ring_size (int): Ring size
            op_str (str): Operation string.
        Returns:
            op (Callable[...,Any]): The operation method for the op_str.
        Raises:
            ValueError : If invalid ring size or op_str is given as input.
        """
        ops = RING_SIZE_TO_OP.get(ring_size, None)

        if ops is None:
            raise ValueError(f"Do not have operations for ring size {ring_size}")

        op = ops.get(op_str, None)
        if op is None:
            raise ValueError(
                f"Operator {op_str} does not exist for ring size {ring_size}"
            )

        return op

    """ TODO: Remove this -- we would use generate_przs since the scenario we are testing is that
    the secret is remotly
    @staticmethod
    def generate_shares(secret, nr_shares, ring_size=2 ** 64):
        from .fixed_precision_tensor import FixedPrecisionTensor

        if not isinstance(secret, (int, FixedPrecisionTensor)):
            secret = FixedPrecisionTensor(value=secret)

        shape = secret.shape
        min_value, max_value = ShareTensor.compute_min_max_from_ring(ring_size)

        generator_shares = np.random.default_rng()

        random_shares = []
        for i in range(nr_shares):
            random_value = generator_shares.integers(
                low=min_value, high=max_value, size=shape
            )
            fpt_value = FixedPrecisionTensor(value=random_value)
            random_shares.append(fpt_value)

        shares_fpt = []
        for i in range(nr_shares):
            if i == 0:
                share = value = random_shares[i]
            elif i < nr_shares - 1:
                share = random_shares[i] - random_shares[i - 1]
            else:
                share = secret - random_shares[i - 1]

            shares_fpt.append(share)

        # Add the ShareTensor class between them
        shares = []
        for rank, share_fpt in enumerate(shares_fpt):
            share_fpt.child = ShareTensor(rank=rank, value=share_fpt.child)
            shares.append(share_fpt)

        return shares
    """

    @staticmethod
    def generate_przs(
        value: Any,
        shape: Tuple[int, ...],
        rank: int,
        parties_info: List[GridURL],
        ring_size: int = 2**32,
        seed_przs: Optional[int] = None,
        generator_przs: Optional[Any] = None,
        init_clients: bool = True,
    ) -> "ShareTensor":

        nr_parties = len(parties_info)

        # Try:
        # 1. First get numpy type if secret is numpy and obtain ring size from there
        # 2. If not get the type from the ring size

        numpy_type = None
        ring_size_final = None

        ring_size_from_type = utils.TYPE_TO_RING_SIZE.get(
            getattr(value, "dtype", None), None
        )
        if ring_size_from_type is None:
            logger.warning("Could not get ring size from {value}")
        else:
            ring_size_final = ring_size_from_type
            numpy_type = value.dtype

        if numpy_type is None:
            numpy_type = utils.RING_SIZE_TO_TYPE.get(ring_size, None)
            ring_size_final = ring_size

        if numpy_type is None:
            raise ValueError(f"Ring size {ring_size} not known how to be treated")

        # relative
        from ..tensor import Tensor

        if (seed_przs is None) == (generator_przs is None):
            raise ValueError("Only seed_przs or generator should be populated")

        if value is None:
            value = Tensor(np.zeros(shape, dtype=numpy_type))

        # TODO: Sending the seed and having each party generate the shares is not safe
        # Since the parties would know some of the other parties shares (this might not impose a risk
        # when shares are not sent between parties -- like private addition/subtraction, but it might
        # impose for multiplication
        # The secret holder should generate the shares and send them to the other parties
        if generator_przs:
            generator_shares = generator_przs
        else:
            generator_shares = np.random.default_rng(seed_przs)

        if isinstance(value.child, ShareTensor):
            value = value.child

        share = ShareTensor(
            value=value.child,
            rank=rank,
            parties_info=parties_info,
            seed_przs=seed_przs,  # type: ignore #TODO:Inspect as we could pass none.
            init_clients=init_clients,
            ring_size=ring_size_final,  # type: ignore
        )

        share.generator_przs = generator_shares
        shares = [
            generator_shares.integers(
                low=share.min_value,
                high=share.max_value,
                size=shape,
                endpoint=True,
                dtype=numpy_type,
            )
            for _ in range(nr_parties)
        ]

        op = ShareTensor.get_op(ring_size_final, "sub")
        przs_share = op(shares[rank], shares[(rank + 1) % nr_parties])
        share.child = op(share.child, przs_share)

        return share

    @staticmethod
    def generate_przs_on_dp_tensor(
        value: Optional[Any],
        shape: Tuple[int],
        rank: int,
        parties_info: List[GridURL],
        seed_przs: int,
        share_wrapper: Any,
        ring_size: int = 2**32,
    ) -> PassthroughTensor:

        if value is not None:
            share = ShareTensor.generate_przs(
                value=value.child,
                shape=shape,
                rank=rank,
                parties_info=parties_info,
                seed_przs=seed_przs,
                ring_size=ring_size,
            )
        else:
            share = ShareTensor.generate_przs(
                value=value,
                shape=shape,
                rank=rank,
                parties_info=parties_info,
                seed_przs=seed_przs,
                ring_size=ring_size,
            )

        share_wrapper.child.child = share

        return share_wrapper

    @staticmethod
    def sanity_check(
        share: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> None:
        """Check type for share

        Args:
            share (Union[int, float, ShareTensor, np.ndarray, torch.Tensor]): value to check

        Raises:
            ValueError: if type is not supported
        """
        if isinstance(share, float):
            raise ValueError("Type float not supported yet!")

        if isinstance(share, np.ndarray) and (
            not np.issubdtype(share.dtype, np.integer)
            and share.dtype != np.dtype("bool")
        ):
            raise ValueError(
                f"NPArray should have type int or bool, but found {share.dtype}"
            )

        if isinstance(share, torch.Tensor) and torch.is_floating_point(share):
            raise ValueError("Torch tensor should have type int, but found float")

    def apply_function(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"], op_str: str
    ) -> "ShareTensor":
        """Apply a given operation.

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): tensor to apply the operator.
            op_str (str): Operator.

        Returns:
            ShareTensor: Result of the operation.
        """

        op = ShareTensor.get_op(self.ring_size, op_str)
        numpy_type = utils.RING_SIZE_TO_TYPE.get(self.ring_size, None)
        if numpy_type is None:
            raise ValueError(f"Do not know numpy type for ring size {self.ring_size}")

        if isinstance(y, ShareTensor):
            utils.get_ring_size(self.ring_size, y.ring_size)
            value = op(self.child, y.child)
        else:
            # TODO: Converting y to numpy because doing "numpy op torch tensor" raises exception
            value = op(self.child, np.array(y, numpy_type))  # TODO: change to np.int64

        res = self.copy_tensor()
        res.child = value
        return res

    def add(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "add" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self + y

        Returns:
            ShareTensor. Result of the operation.
        """

        ShareTensor.sanity_check(y)

        new_share = self.apply_function(y, "add")

        return new_share

    def sub(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "sub" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self - y

        Returns:
            ShareTensor. Result of the operation.
        """

        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "sub")
        return new_share

    def rsub(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "rsub" operation between "self" and "y"

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): y - self

        Returns:
            ShareTensor. Result of the operation.
        """

        ShareTensor.sanity_check(y)
        new_self = self.mul(-1)
        new_share = new_self.apply_function(y, "add")
        return new_share

    def mul(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "mul" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self * y

        Returns:
            ShareTensor. Result of the operation.
        """
        # if isinstance(y, ShareTensor):
        #     raise ValueError(
        #         "We should not reach this point for private multiplication. Only public one"
        #     )

        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "mul")
        return new_share

    def matmul(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "matmul" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self @ y.

        Returns:
            ShareTensor: Result of the operation.
        """
        if isinstance(y, ShareTensor):
            raise ValueError("Private matmul not supported yet")

        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "matmul")
        return new_share

    def rmatmul(self, y: torch.Tensor) -> "ShareTensor":
        """Apply the "rmatmul" operation between "y" and "self".

        Args:
            y (torch.Tensor): y @ self

        Returns:
            ShareTensor. Result of the operation.
        """
        if isinstance(y, ShareTensor):
            raise ValueError("Private matmul not supported yet")

        ShareTensor.sanity_check(y)
        new_share = y.apply_function(self, "matmul")
        return new_share

    def lt(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "lt" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self < y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "lt")
        return new_share

    def gt(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "gt" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self > y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "gt")
        return new_share

    def ge(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "ge" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self >= y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "ge")
        return new_share

    def le(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "le" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self <= y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "le")
        return new_share

    def ne(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "ne" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self != y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        ShareTensor.sanity_check(y)
        new_share = self.apply_function(y, "ne")
        return new_share

    def bit_decomposition(self) -> "ShareTensor":
        """Apply the "decomposition" operation on self

        Args:
            None

        Returns:
            ShareTensor. Result of the operation.
        """
        raise ValueError(
            "It should not reach this point since we generate SMPCAction for this"
        )

    def eq(self, other: Any) -> bool:
        """Equal operator.
        Check if "self" is equal with another object given a set of
            attributes to compare.
        Args:
            other (Any): Value to compare.
        Returns:
            bool: True if equal False if not.
        """
        # TODO: Rasswanth: Fix later after the comparison operation
        # relative
        # from .... import Tensor

        # if (
        #     isinstance(self.child, Tensor)
        #     and isinstance(other.child, Tensor)
        #     and (self.child != other.child).child.any()  # type: ignore
        # ):
        #     return False

        # if (
        #     isinstance(self.child, np.ndarray)
        #     and isinstance(other.child, np.ndarray)
        #     and (self.child != other.child).any()
        # ):
        #     return False

        # if self.rank != other.rank:
        #     return False

        # if self.ring_size != other.ring_size:
        #     return False

        # if self.nr_parties != other.nr_parties:
        #     return False

        # return True

        return self.child == other.child

    # TRASK: commenting out because ShareTEnsor doesn't appear to have .session_uuid or .config
    # def div(
    #     self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    # ) -> "ShareTensor":
    #     """Apply the "div" operation between "self" and "y".
    #
    #     Args:
    #         y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): Denominator.
    #
    #     Returns:
    #         ShareTensor: Result of the operation.
    #
    #     Raises:
    #         ValueError: If y is not an integer or LongTensor.
    #     """
    #     if not isinstance(y, (int, torch.LongTensor)):
    #         raise ValueError("Div works (for the moment) only with integers!")
    #
    #     res = ShareTensor(session_uuid=self.session_uuid, config=self.config)
    #     # res = self.apply_function(y, "floordiv")
    #     res.tensor = self.tensor // y
    #     return res

    def bit_extraction(self, pos: int = 0) -> ShareTensor:
        """Extracts the bit at the specified position.

        Args:
            pos (int): position to extract bit.

        Returns:
            ShareTensor : extracted bits at specific position.

        Raises:
            ValueError: If invalid position is provided.
        """
        ring_bits = utils.get_nr_bits(self.ring_size)
        if pos < 0 or pos > ring_bits - 1:
            raise ValueError(
                f"Invalid position for bit_extraction: {pos}, must be in range:[0,{ring_bits-1}]"
            )
        shape = self.shape
        numpy_type = utils.RING_SIZE_TO_TYPE[self.ring_size]
        # logical shift
        bit_mask = np.ones(shape, dtype=numpy_type) << pos
        value = self.child & bit_mask
        value = value.astype(np.bool_)
        share = self.copy_tensor()
        share.child = value
        return share

    @staticmethod
    def hook_method(__self: ShareTensor, method_name: str) -> Callable[..., Any]:
        """Hook a framework method.

        Args:
            method_name (str): method to hook

        Returns:
            A hooked method
        """

        def method_all_shares(
            _self: ShareTensor, *args: List[Any], **kwargs: Dict[Any, Any]
        ) -> Any:

            share = _self.child
            if method_name != "resize":
                method = getattr(share, method_name)
            else:
                # Should be modified to remove copy
                # https://stackoverflow.com/questions/23253144/numpy-the-array-doesnt-have-its-own-data
                share = share.copy()
                method = getattr(share, method_name)

            if method_name not in INPLACE_OPS:
                new_share = method(*args, **kwargs)
            else:
                method(*args, **kwargs)
                new_share = share

            res = _self.copy_tensor()

            # TODO : Some operations return np.int64 by default, should modify
            # when we have support for np.int64 or do explicit casting.
            if method_name == "trace":
                new_share = np.array(new_share, dtype=np.int32)
            res.child = new_share

            return res

        return functools.partial(method_all_shares, __self)

    def __getattribute__(self, attr_name: str) -> Any:
        if attr_name in METHODS_FORWARD_ALL_SHARES or attr_name in INPLACE_OPS:
            return ShareTensor.hook_method(self, attr_name)

        return object.__getattribute__(self, attr_name)

    def _object2proto(self) -> ShareTensor_PB:
        proto_init_kwargs = {
            "rank": self.rank,
            "parties_info": [serialize(party) for party in self.parties_info],
            "seed_przs": self.seed_przs,
            "ring_size": sy.serialize(self.ring_size, to_bytes=True),
        }
        if isinstance(self.child, np.ndarray):
            proto_init_kwargs["array"] = serialize(self.child)
        elif isinstance(self.child, torch.Tensor):
            proto_init_kwargs["array"] = serialize(np.array(self.child))
        else:
            proto_init_kwargs["tensor"] = serialize(self.child)

        return ShareTensor_PB(**proto_init_kwargs)

    @staticmethod
    def _proto2object(proto: ShareTensor_PB) -> "ShareTensor":
        init_kwargs = {
            "rank": proto.rank,
            "parties_info": [deserialize(party) for party in proto.parties_info],
            "seed_przs": proto.seed_przs,
            "ring_size": int(sy.deserialize(proto.ring_size, from_bytes=True)),
        }
        if proto.HasField("tensor"):
            init_kwargs["value"] = deserialize(proto.tensor)
        else:
            init_kwargs["value"] = deserialize(proto.array)

        # init_kwargs["init_clients"] = True
        res = ShareTensor(**init_kwargs)
        generator_przs = np.random.default_rng(proto.seed_przs)
        res.generator_przs = generator_przs
        return res

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return ShareTensor_PB

    __add__ = add
    __radd__ = add
    __sub__ = sub
    __rsub__ = rsub
    __mul__ = mul
    __rmul__ = mul
    __matmul__ = matmul
    __rmatmul__ = rmatmul
    __lt__ = lt
    __gt__ = gt
    __ge__ = ge
    __le__ = le
    __eq__ = eq
    __ne__ = ne



"""Specific PyGrid exceptions."""


class PyGridError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class PermissionsNotDefined(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = "This message hasn't permissions defined, Please set up message permissions."
        super().__init__(message)


class BadPayloadException(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = (
                "Missing mandatory message fields or wrong message format and type."
            )
        super().__init__(message)


class AuthorizationError(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = "User is not authorized for this operation!"
        super().__init__(message)


class OwnerAlreadyExistsError(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = "This PyGrid domain already has an owner!"
        super().__init__(message)


class RoleNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Role ID not found!"
        super().__init__(message)


class ModelNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Model ID not found!"
        super().__init__(message)


class CycleNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Cycle not found!"
        super().__init__(message)


class PlanNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Plan ID not found!"
        super().__init__(message)


class ProcessNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Not found any process related with this cycle and worker ID."
        super().__init__(message)


class PlanInvalidError(PyGridError):
    def __init__(self) -> None:
        message = "Plan is not valid"
        super().__init__(message)


class PlanTranslationError(PyGridError):
    def __init__(self) -> None:
        message = "Failed to translate a Plan"
        super().__init__(message)


class WorkerNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Worker ID not found!"
        super().__init__(message)


class ProtocolNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Protocol ID not found!"
        super().__init__(message)


class FLProcessConflict(PyGridError):
    def __init__(self) -> None:
        message = "FL Process already exists."
        super().__init__(message)


class MaxCycleLimitExceededError(PyGridError):
    def __init__(self) -> None:
        message = "There are no cycles remaining"
        super().__init__(message)


class UserNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "User not found!"
        super().__init__(message)


class EnvironmentNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Environment not found!"
        super().__init__(message)


class SetupNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Setup not found!"
        super().__init__(message)


class GroupNotFoundError(PyGridError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidRequestKeyError(PyGridError):
    def __init__(self) -> None:
        message = "Invalid request key!"
        super().__init__(message)


class InvalidCredentialsError(PyGridError):
    def __init__(self) -> None:
        message = "Invalid credentials!"
        super().__init__(message)


class MissingRequestKeyError(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = "Missing request key!"
        super().__init__(message)


class AssociationRequestError(PyGridError):
    def __init__(self) -> None:
        message = "Association Request ID not found!"
        super().__init__(message)


class AssociationError(PyGridError):
    def __init__(self) -> None:
        message = "Association ID not found!"
        super().__init__(message)


class RequestError(PyGridError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class DatasetNotFoundError(PyGridError):
    def __init__(self) -> None:
        message = "Dataset ID not found!"
        super().__init__(message)


class InvalidParameterValueError(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = "Passed paramater value not valid!"
        super().__init__(message)


class AppInSleepyMode(PyGridError):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = (
                "This app is in sleep mode. Please undergo the initial setup first"
            )
        super().__init__(message)



# TODO: move this service back to PyGrid where it belongs

# # stdlib
# from datetime import datetime
# import os
# from typing import List
# from typing import Type
# from typing import Union
#
# # third party
# from nacl.encoding import HexEncoder
# from nacl.signing import VerifyKey
#
# # syft absolute
# from syft import serialize
# from syft.core.common.message import ImmediateSyftMessageWithReply
# from syft.core.node.abstract.node import AbstractNode
# from syft.core.node.common.service.auth import service_auth
# from syft.core.node.common.service.node_service import ImmediateNodeServiceWithReply
# from syft.grid.client.client import connect
# from syft.grid.client.grid_connection import GridHTTPConnection
# from .infra_manager_messages import CreateWorkerMessage
# from .infra_manager_messages import CreateWorkerResponse
# from .infra_manager_messages import DeleteWorkerMessage
# from .infra_manager_messages import DeleteWorkerResponse
# from .infra_manager_messages import GetWorkerInstanceTypesMessage
# from .infra_manager_messages import GetWorkerInstanceTypesResponse
# from .infra_manager_messages import GetWorkerMessage
# from .infra_manager_messages import GetWorkerResponse
# from .infra_manager_messages import GetWorkersMessage
# from .infra_manager_messages import GetWorkersResponse
#
# # relative
# from ....core.infrastructure import AWS_Serverfull
# from ....core.infrastructure import AZURE
# from ....core.infrastructure import Config
# from ....core.infrastructure import GCP
# from ....core.infrastructure import Provider
# from ....core.infrastructure import aws_utils
# from ....core.infrastructure import azure_utils
# from ....core.infrastructure import gcp_utils
# from ...database.tables.environment import states
# from ...database.utils import model_to_json
# from ...exceptions import AuthorizationError
# from ...exceptions import MissingRequestKeyError
#
# # TODO: Modify existing routes or add new ones, to
# # 1. allow admin to get all workers deployed by a specific user
# # 2. allow admin to get all workers deployed by all users
#
# SUPPORTED_PROVIDERS = [
#     "aws",
#     "azure",
#     "gcp",
# ]  # todo: add azure and gcp after testing worker deployment
# PROVIDER_UTILS = {
#     "aws": aws_utils,
#     "azure": azure_utils,
#     "gcp": gcp_utils,
# }
#
#
# def get_worker_instance_types_msg(
#     msg: GetWorkerInstanceTypesMessage, node: AbstractNode, verify_key: VerifyKey
# ) -> GetWorkerInstanceTypesResponse:
#     try:
#         _current_user_id = msg.content.get("current_user", None)
#         provider = os.environ.get("CLOUD_PROVIDER")
#         region = os.environ.get("REGION")
#
#         if provider not in SUPPORTED_PROVIDERS:
#             raise Exception("Provider not supported")
#
#         # todo: We can make worker deployment a permissible operation
#         # Users can deploy certain instance types (such as "Free Tier Instances") without permission
#         # But to deploy other instance types, example those which are costly, they would need to ask permission
#         # This servide would then return only those instance types which the users has the permission to deploy
#
#         if provider == "aws":
#             _msg = PROVIDER_UTILS[provider].get_all_instance_types(region)
#         elif provider == "gcp":
#             zone = os.environ.get("ZONE", None)
#             _msg = PROVIDER_UTILS[provider].get_all_instance_types(zone=zone)
#         elif provider == "azure":
#             location = os.environ.get("location", None)
#             _msg = PROVIDER_UTILS[provider].get_all_instance_types(location=location)
#         return GetWorkerInstanceTypesResponse(
#             address=msg.reply_to, status_code=200, content=_msg
#         )
#     except Exception as e:
#         return GetWorkerInstanceTypesResponse(
#             address=msg.reply_to, status_code=500, content={"error": str(e)}
#         )
#
#
# def create_worker_msg(
#     msg: CreateWorkerMessage, node: AbstractNode, verify_key: VerifyKey
# ) -> CreateWorkerResponse:
#     try:
#         _current_user_id = msg.content.get("current_user", None)
#         instance_type = msg.content.get("instance_type", None)
#         _worker_port = msg.content.get("port", 5001)
#
#         users = node.users
#
#         if not _current_user_id:
#             _current_user_id = users.first(
#                 verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
#             ).id
#
#         if instance_type is None:
#             raise MissingRequestKeyError
#
#         config = Config(
#             app=Config(name="worker", count=1, id=len(node.environments.all()) + 1),
#             apps=[Config(name="worker", count=1, port=_worker_port)],
#             serverless=False,
#             websockets=False,
#             provider=os.getenv("CLOUD_PROVIDER", "AWS"),
#             ##TODO(amr): encapsulate each cloud provider to config to aws, azure, gcp
#             vpc=Config(
#                 region=os.getenv("REGION", None),
#                 instance_type=Config(InstanceType=instance_type),
#             ),
#             azure=Config(
#                 location=os.getenv("location", None),
#                 subscription_id=os.getenv("subscription_id", None),
#                 client_id=os.getenv("client_id", None),
#                 client_secret=os.getenv("client_secret", None),
#                 tenant_id=os.getenv("tenant_id", None),
#                 vm_size=instance_type,
#             ),
#             gcp=Config(
#                 project_id=os.getenv("project_id", None),
#                 region=os.getenv("region", None),
#                 zone=os.getenv("zone", None),
#                 machine_type=instance_type,
#             ),
#         )
#
#         deployment = None
#         deployed = False
#
#         if config.provider == "aws":
#             deployment = AWS_Serverfull(config=config)
#         elif config.provider == "azure":
#             deployment = AZURE(config=config)
#         elif config.provider == "gcp":
#             deployment = GCP(config=config)
#
#         if deployment.validate():
#             deployed, output = deployment.deploy()  # Deploy
#             if deployed:
#                 env_parameters = {
#                     "id": config.app.id,
#                     "provider": config.provider,
#                     "created_at": datetime.now(),
#                     "state": states["success"],
#                     "address": output["instance_0_endpoint"]["value"][0]
#                     + ":"
#                     + str(_worker_port),
#                 }
#
#                 if config.provider == "aws":
#                     env_parameters["region"] = config.vpc.region
#                     env_parameters[
#                         "instance_type"
#                     ] = config.vpc.instance_type.InstanceType
#                 elif config.provider == "azure":
#                     env_parameters["region"] = config.azure.location
#                     env_parameters["instance_type"] = config.azure.vm_size
#                 elif config.provider == "gcp":
#                     env_parameters["region"] = config.gcp.region
#                     env_parameters["instance_type"] = config.gcp.machine_type
#
#                 new_env = node.environments.register(**env_parameters)
#                 node.environments.association(
#                     user_id=_current_user_id, env_id=new_env.id
#                 )
#             else:
#                 node.environments.set(id=config.app.id, state=states["failed"])
#                 raise Exception("Worker creation failed!")
#         final_msg = "Worker created successfully!"
#         return CreateWorkerResponse(
#             address=msg.reply_to, status_code=200, content={"message": final_msg}
#         )
#     except Exception as e:
#         return CreateWorkerResponse(
#             address=msg.reply_to, status_code=500, content={"error": str(e)}
#         )
#
#
# def get_worker_msg(
#     msg: GetWorkerMessage, node: AbstractNode, verify_key: VerifyKey
# ) -> GetWorkerResponse:
#     try:
#
#         worker_id = msg.content.get("worker_id", None)
#         _current_user_id = msg.content.get("current_user", None)
#
#         users = node.users
#
#         if not _current_user_id:
#             _current_user_id = users.first(
#                 verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
#             ).id
#
#         env_ids = [
#             env.id for env in node.environments.get_environments(user=_current_user_id)
#         ]
#         is_admin = users.can_manage_infrastructure(user_id=_current_user_id)
#
#         if (int(worker_id) in env_ids) or is_admin:
#             worker = node.environments.first(id=int(worker_id))
#
#             try:
#                 worker_client = connect(
#                     url="http://" + worker.address,
#                     conn_type=GridHTTPConnection,  # HTTP Connection Protocol
#                 )
#
#                 node.environments.set(
#                     id=worker.id,
#                     syft_address=serialize(worker_client.address)
#                     .SerializeToString()
#                     .decode("ISO-8859-1"),
#                 )
#
#                 node.in_memory_client_registry[worker_client.domain_id] = worker_client
#             except Exception as e:
#                 return GetWorkerResponse(
#                     address=msg.reply_to,
#                     status_code=500,
#                     content={"error": str(e)},
#                 )
#             _msg = model_to_json(node.environments.first(id=int(worker_id)))
#         else:
#             _msg = {}
#
#         return GetWorkerResponse(address=msg.reply_to, status_code=200, content=_msg)
#     except Exception as e:
#         return GetWorkerResponse(
#             address=msg.reply_to, status_code=500, content={"error": str(e)}
#         )
#
#
# def get_workers_msg(
#     msg: GetWorkersMessage, node: AbstractNode, verify_key: VerifyKey
# ) -> GetWorkersResponse:
#     try:
#         _current_user_id = msg.content.get("current_user", None)
#         include_all = msg.content.get("include_all", True)
#         include_failed = msg.content.get("include_failed", False)
#         include_destroyed = msg.content.get("include_destroyed", False)
#
#         if not _current_user_id:
#             _current_user_id = node.users.first(
#                 verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
#             ).id
#
#         envs = node.environments.get_environments(user=_current_user_id)
#
#         workers = []
#         print("Node environments: ", node.environments.all()[0].id)
#         for env in envs:
#             print("Here!", env.id)
#             _env = node.environments.first(id=env.id)
#
#             if (
#                 include_all
#                 or (_env.state == states["success"])
#                 or (include_failed and _env.state == states["failed"])
#                 or (include_destroyed and _env.state == states["destroyed"])
#             ):
#                 worker = model_to_json(_env)
#                 del worker["syft_address"]
#                 workers.append(worker)
#
#         _msg = workers
#
#         return GetWorkersResponse(address=msg.reply_to, status_code=200, content=_msg)
#     except Exception as e:
#         return GetWorkersResponse(
#             address=msg.reply_to, status_code=False, content={"error": str(e)}
#         )
#
#
# def del_worker_msg(
#     msg: DeleteWorkerMessage, node: AbstractNode, verify_key: VerifyKey
# ) -> DeleteWorkerResponse:
#     try:
#         # Get Payload Content
#         worker_id = msg.content.get("worker_id", None)
#         _current_user_id = msg.content.get("current_user", None)
#
#         users = node.users
#         if not _current_user_id:
#             _current_user_id = users.first(
#                 verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
#             ).id
#
#         is_admin = users.can_manage_infrastructure(user_id=_current_user_id)
#
#         envs = [
#             int(env.id)
#             for env in node.environments.get_environments(user=_current_user_id)
#         ]
#         created_by_current_user = int(worker_id) in envs
#
#         # Owner / Admin
#         if not is_admin and not created_by_current_user:
#             raise AuthorizationError("You're not allowed to delete this worker!")
#
#         env = node.environments.first(id=worker_id)
#         _config = Config(provider=env.provider, app=Config(name="worker", id=worker_id))
#
#         if env.state == states["success"]:
#             worker_dir = os.path.join(
#                 "/home/ubuntu/.pygrid/apps/aws/workers/", str(worker_id)
#             )
#             success = Provider(worker_dir).destroy()
#             if success:
#                 node.environments.set(
#                     id=worker_id, state=states["destroyed"], destroyed_at=datetime.now()
#                 )
#
#         if env.state == states["destroyed"]:
#             return DeleteWorkerResponse(
#                 address=msg.reply_to,
#                 status_code=200,
#                 content={"message": "Worker was deleted successfully!"},
#             )
#         else:
#             raise Exception("Worker deletion failed")
#
#     except Exception as e:
#         return DeleteWorkerResponse(
#             address=msg.reply_to, status_code=False, content={"error": str(e)}
#         )
#
#
# class DomainInfrastructureService(ImmediateNodeServiceWithReply):
#
#     msg_handler_map = {
#         GetWorkerInstanceTypesMessage: get_worker_instance_types_msg,
#         CreateWorkerMessage: create_worker_msg,
#         GetWorkerMessage: get_worker_msg,
#         GetWorkersMessage: get_workers_msg,
#         DeleteWorkerMessage: del_worker_msg,
#     }
#
#     @staticmethod
#     @service_auth(guests_welcome=True)
#     def process(
#         node: AbstractNode,
#         msg: Union[
#             GetWorkerInstanceTypesMessage,
#             CreateWorkerMessage,
#             GetWorkerMessage,
#             GetWorkersMessage,
#             DeleteWorkerMessage,
#         ],
#         verify_key: VerifyKey,
#     ) -> Union[
#         GetWorkerInstanceTypesResponse,
#         CreateWorkerResponse,
#         GetWorkerResponse,
#         GetWorkersResponse,
#         DeleteWorkerResponse,
#     ]:
#         return DomainInfrastructureService.msg_handler_map[type(msg)](
#             msg=msg, node=node, verify_key=verify_key
#         )
#
#     @staticmethod
#     def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
#         return [
#             GetWorkerInstanceTypesMessage,
#             CreateWorkerMessage,
#             # CheckWorkerDeploymentMessage,
#             # UpdateWorkerMessage,
#             GetWorkerMessage,
#             GetWorkersMessage,
#             DeleteWorkerMessage,
#         ]



# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# relative
from ...... import deserialize
from ...... import serialize
from ......proto.core.node.domain.service.request_answer_message_pb2 import (
    RequestAnswerMessage as RequestAnswerMessage_PB,
)
from ......proto.core.node.domain.service.request_answer_response_pb2 import (
    RequestAnswerResponse as RequestAnswerResponse_PB,
)
from .....common import UID
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....io.address import Address
from ..request_receiver.request_receiver_messages import RequestStatus


@serializable()
class RequestAnswerMessage(ImmediateSyftMessageWithReply):
    __slots__ = ["request_id"]

    def __init__(self, request_id: UID, reply_to: Address, address: Address):
        super().__init__(reply_to, address)
        self.request_id = request_id

    def _object2proto(self) -> RequestAnswerMessage_PB:
        msg = RequestAnswerMessage_PB()
        msg.request_id.CopyFrom(serialize(obj=self.request_id))
        msg.address.CopyFrom(serialize(obj=self.address))
        msg.reply_to.CopyFrom(serialize(obj=self.reply_to))
        return msg

    @staticmethod
    def _proto2object(proto: RequestAnswerMessage_PB) -> "RequestAnswerMessage":
        return RequestAnswerMessage(
            request_id=deserialize(blob=proto.request_id),
            address=deserialize(blob=proto.address),
            reply_to=deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RequestAnswerMessage_PB


@serializable()
class RequestAnswerResponse(ImmediateSyftMessageWithoutReply):

    __slots__ = ["status", "request_id"]

    def __init__(self, status: RequestStatus, request_id: UID, address: Address):
        super().__init__(address)
        self.status = status
        self.request_id = request_id

    def _object2proto(self) -> RequestAnswerResponse_PB:
        msg = RequestAnswerResponse_PB()
        msg.request_id.CopyFrom(serialize(obj=self.request_id))
        msg.status = self.status.value
        msg.address.CopyFrom(serialize(obj=self.address))
        return msg

    @staticmethod
    def _proto2object(proto: RequestAnswerResponse_PB) -> "RequestAnswerResponse":
        request_response = RequestAnswerResponse(
            status=RequestStatus(proto.status),
            request_id=deserialize(blob=proto.request_id),
            address=deserialize(blob=proto.address),
        )
        return request_response

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RequestAnswerResponse_PB



# future
from __future__ import annotations

# stdlib
from collections.abc import KeysView
from typing import List
from typing import Optional

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

# relative
from ....adp.entity import Entity
from ....adp.idp_gaussian_mechanism import iDPGaussianMechanism
from ....common.serde import _serialize
from ..node_table.entity import Entity as EntitySchema
from ..node_table.ledger import Ledger
from ..node_table.mechanism import Mechanism as MechanismSchema
from ..node_table.user import SyftUser
from .database_manager import DatabaseManager


class EntityManager(DatabaseManager):
    schema = EntitySchema

    def __init__(self, database: Engine) -> None:
        super().__init__(db=database, schema=EntityManager.schema)

    def register(self, name: str):  # type: ignore
        entity = Entity(name)  # Constructor
        _obj: EntitySchema = self._schema(name=entity.name)  # type: ignore
        _obj.obj = entity
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.add(_obj)
        session_local.commit()
        session_local.close()
        return entity.name


class MechanismManager(DatabaseManager):
    schema = MechanismSchema

    def __init__(self, database: Engine) -> None:
        super().__init__(db=database, schema=MechanismManager.schema)

    def register(self, mechanism):  # type: ignore
        _obj = self._schema(
            entity_name=mechanism.entity_name,
            user_key=_serialize(mechanism.user_key, to_bytes=True),
        )
        _obj.obj = mechanism
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.add(_obj)
        session_local.commit()
        obj_id = _obj.id
        session_local.close()
        return obj_id


class AbstractLedger:
    def register_mechanisms(self, mechanisms: list) -> None:
        raise NotImplementedError

    def query(  # type: ignore
        self, entity_name: Optional[str] = None, user_key: Optional[str] = None
    ) -> List[iDPGaussianMechanism]:
        raise NotImplementedError

    def keys(self) -> KeysView:
        raise NotImplementedError

    def get_user_budget(self, user_key: VerifyKey) -> float:
        raise NotImplementedError


class DictLedger(AbstractLedger):
    def __init__(self) -> None:
        super().__init__()
        self._entity_dict: dict[str, iDPGaussianMechanism] = {}
        self._user_key_dict: dict[VerifyKey, iDPGaussianMechanism] = {}
        self._user_budget: dict[VerifyKey, float] = {}

    def register_mechanisms(self, mechanisms: list) -> None:
        for m in mechanisms:
            # if entity doesnt exist:
            if m.entity_name not in self._entity_dict:
                # write to the entity table
                self._entity_dict[m.entity_name] = m
            if m.user_key not in self._user_key_dict:
                self._user_key_dict[m.user_key] = m

    def query(
        self, entity_name: Optional[str] = None, user_key: Optional[VerifyKey] = None
    ) -> List[iDPGaussianMechanism]:
        mechanisms = []
        if entity_name is not None and user_key is None:
            if entity_name in self._entity_dict:
                mechanisms.append(self._entity_dict[entity_name])
                return mechanisms
        elif user_key is not None and entity_name is None:
            if user_key in self._user_key_dict:
                mechanisms.append(self._user_key_dict[user_key])
                return mechanisms
        elif entity_name is not None and user_key is not None:
            mechanisms.extend(self._entity_dict.values())
            m_set = set(mechanisms)
            return list(m_set.intersection(self._user_key_dict.values()))
        else:
            mechanisms.extend(self._entity_dict.values())
            mechanisms.extend(self._user_key_dict.values())
            return list(set(mechanisms))

        return mechanisms

    def keys(self) -> KeysView:
        return self._entity_dict.keys()

    def get_user_budget(self, user_key: VerifyKey) -> float:
        if user_key in self._user_budget:
            return self._user_budget[user_key]
        default_budget = 999999999
        print(f"Warning, no user found returning default budget. {default_budget}")
        return float(default_budget)


class DatabaseLedger(DatabaseManager, AbstractLedger):
    schema = Ledger

    def __init__(self, database: Engine) -> None:
        super().__init__(db=database, schema=DatabaseLedger.schema)
        self.entity_manager = EntityManager(database)
        self.mechanism_manager = MechanismManager(database)

    def get_user_budget(self, user_key: VerifyKey) -> float:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        serialized_key = user_key.encode(encoder=HexEncoder).decode("utf-8")
        budget = (
            session_local.query(SyftUser)
            .filter_by(verify_key=serialized_key)
            .first()
            .budget
        )

        return budget

    def keys(self) -> KeysView:
        return KeysView(self.entity_manager.all())  # type: ignore

    def register_mechanisms(self, mechanisms: list) -> None:
        for m in mechanisms:
            # if entity doesnt exist:
            if not self.entity_manager.contain(name=m.entity_name):
                # write to the entity table
                self.entity_manager.register(name=m.entity_name)
            self.mechanism_manager.register(m)  # type: ignore

    def query(  # type: ignore
        self, entity_name: Optional[str] = None, user_key: Optional[str] = None
    ) -> List[iDPGaussianMechanism]:
        if entity_name is not None and user_key is not None:
            return self.mechanism_manager.query(
                entity_name=entity_name, user_key=user_key
            )
        elif entity_name is not None and user_key is None:
            return self.mechanism_manager.query(entity_name=entity_name)
        elif entity_name is None and user_key is not None:
            return self.mechanism_manager.query(user_key=user_key)
        else:
            return self.mechanism_manager.all()



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import numpy as np

# relative
from . import array  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..misc.union import UnionGenerator
from ..util import generic_update_ast

LIB_NAME = "numpy"
PACKAGE_SUPPORT = {"lib": LIB_NAME}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("numpy", np),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("numpy.ndarray", "numpy.ndarray", np.ndarray),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        ("numpy.ndarray.shape", "syft.lib.python.Tuple"),
        ("numpy.ndarray.strides", "syft.lib.python.Tuple"),
        ("numpy.ndarray.ndim", "syft.lib.python.Int"),
        ("numpy.ndarray.size", "syft.lib.python.Int"),
        ("numpy.ndarray.itemsize", "syft.lib.python.Int"),
        ("numpy.ndarray.nbytes", "syft.lib.python.Int"),
        # ("numpy.ndarray.dtype", "numpy.ndarray"), # SECURITY WARNING: DO NOT ADD TO ALLOW LIST YET
        ("numpy.ndarray.T", "numpy.ndarray"),
        # ("numpy.ndarray.real", "numpy.ndarray"), # requires dtype complex
        # ("numpy.ndarray.imag", "numpy.ndarray"), # requires dtype complex
        # Array methods - Array Conversion
        (
            "numpy.ndarray.item",
            UnionGenerator[
                "syft.lib.python.Bool", "syft.lib.python.Float", "syft.lib.python.Int"
            ],
        ),
        ("numpy.ndarray.byteswap", "numpy.ndarray"),
        ("numpy.ndarray.copy", "numpy.ndarray"),
        ("numpy.ndarray.view", "numpy.ndarray"),
        ("numpy.ndarray.__add__", "numpy.ndarray"),
        ("numpy.ndarray.sum", "numpy.ndarray"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



"""

***************************************************
Tutorial: Serialization and Deserialization in Syft
***************************************************

In this file, we have the main Serializable class which orchestrates
the serialization of objects within the Syft ecosystem. Users and
developers of Syft need to serialize objects for a myriad of reasons,
but the most common 3 are:

- to save an object into a database which requires serialization (such as key-value dbs like Redis)
- to send an object over the network (any protocol).
- to save an object to disk.

All serialization in Syft uses a library called 'protobuf'.
This is a technology developed at Google for fast, secure
serialization of objects (https://developers.google.com/protocol-buffers).
We use an existing library like Protobuf for multiple reasons:

- Protobuf creates cross-language serialization abilities
- Protobuf is secure
- Protobuf is fast.

In short, lots of projects need serialization and so a lot of time
and effort has gone into creating great serialization libraries. Thus,
we want to inherit the work of others instead of having to reinvest the time
to build these things ourselves.

.. note:: DO NOT attempt to create your own serialization by
    creating strings out of objects yourself. Not only will this almost
    certainly, be slower than using technology like protobuf,
    but you will require everyone to re-implement your serialization
    techniques in every language which needs to support your object
    AND you run the risk of introducing dangerous SECURITY FLAWS.
    This is a place where we definitely want to use a robust library
    like protobuf.


Serializing and Deserializing Syft Objects:
###########################################

If you want to serialize an object in the syft ecosystem, the easiest way
to do so is to just call one of the serialization methods which we install
on the object for your convenience::

    import syft as sy

    # ObjectWithID is the simplest object in the Syft ecosystem
    # It's the parent class of many other classes. As the name
    # implies it's literally just an object with an ID.
    # We'll use it to show serialization, but you can use the
    # same approach with any serializable object in the Syft
    # ecosystem.
    from syft.core.common.object import ObjectWithID

    # this creates an object which has an id
    my_object = ObjectWithID()

    print(my_object)
    # >>> <ObjectWithID:fb1bb067-5bb7-4c49-bece-e700ab0a1514>

    # by default, serialize() will serialize it to a protobuf Message object
    proto_obj = serialize(my_object)

    print(proto_obj)
    # >>> obj_type: "syft.core.common.object.ObjectWithID"
    # >>> id {
    # >>>   obj_type: "syft.core.common.uid.UID"
    # >>>   value: "23hi23hgo2ih23ih2;o3igh;2oih;iagapwihpag"
    # >>> }

    # deserialization also assumes you are accepting a protobuf object
    my_object_again = sy.deserialize(blob=proto_obj)

    print(my_object_again)
    # >>> <ObjectWithID:fb1bb067-5bb7-4c49-bece-e700ab0a1514>

However, perhaps the best thing about protobuf is that it can easily
be turned into a wide variety of very portable representations. We have
convenience functions for 4 popular representations: protobuf, json, binary,
and hex. Setup::

    import syft as sy
    from syft.core.common.object import ObjectWithID

    obj = ObjectWithId()

Protobuf
********

.. code::

    proto_obj = serialize(obj, to_proto=True)
    proto_obj = serialize(obj, to_proto=True)
    proto_obj = serialize(obj, to_proto=True)

    print(proto_obj)
    # >>> obj_type: "syft.core.common.object.ObjectWithID"
    # >>> id {
    # >>>   obj_type: "syft.core.common.uid.UID"
    # >>>   value: "23hi23hgo2ih23ih2;o3igh;2oih;iagapwihpag"
    # >>> }

    obj_again = sy.deserialize(blob=proto_obj, from_proto=True)

Binary
******

.. code::

    binary_obj = serialize(obj, to_bytes=True)
    binary_obj = serialize(obj, to_bytes=True)
    binary_obj = serialize(obj, to_bytes=True)

    # print(binary_obj)
    # >>> b'{  "objType": "syft.core.common.object.ObjectWithID",
    # >>> "id": {    "objType": "syft.core.common.uid.UID",
    # >>> "value": "+xuwZ1u3TEm+zucAqwoVFA=="  }}'

    obj_again = sy.deserialize(blob=proto_obj, from_bytes=True)

Now we can continue with the class definition for the Serializable class, which
is the parent class for all serializable objects within Syft.

If you'd like to see a simple example of a class which can be serialized, please read
the source code of :py:mod:`syft.core.common.object.ObjectWithID`.
"""

# relative
from .deserialize import _deserialize  # noqa: F401
from .serialize import _serialize  # noqa: F401



# relative
from .abs import AbsOp  # noqa: 401
from .add import AddOp  # noqa: 401
from .argmax import ArgMaxOp  # noqa: 401
from .argmin import ArgMinOp  # noqa: 401
from .argsort import ArgSortOp  # noqa: 401
from .copy import CopyOp  # noqa: 401
from .dot import DotOp  # noqa: 401
from .getitem import GetItemOp  # noqa: 401
from .invert import InvertOp  # noqa: 401
from .mul import MulOp  # noqa: 401
from .op import Op  # noqa: 401
from .pow import PowOp  # noqa: 401
from .repeat import RepeatOp  # noqa: 401
from .reshape import ReshapeOp  # noqa: 401
from .resize import ResizeOp  # noqa: 401
from .rpow import RPowOp  # noqa: 401
from .rshift import RShiftOp  # noqa: 401
from .sort import SortOp  # noqa: 401
from .sub import SubOp  # noqa: 401
from .sum import SumOp  # noqa: 401
from .transpose import TransposeOp  # noqa: 401
from .truediv import TrueDivOp  # noqa: 401



# stdlib
from typing import Tuple
from uuid import UUID

# third party
from numpy import ndarray

# relative
from ..tensor import AutogradTensor
from .op import Op


class TransposeOp(Op):
    """Repeat operation across a dimension"""

    def forward(self, x: AutogradTensor, *dims: Tuple[int]) -> AutogradTensor:  # type: ignore
        self.x = x
        self.dims = dims

        reverse_t_dims = {}
        for i, d in enumerate(self.dims):
            reverse_t_dims[d] = i

        l_dims = sorted(
            [(x[0], x[1]) for x in reverse_t_dims.items()], key=lambda x: x[0]
        )
        self.reverse_dims = [x[1] for x in l_dims]

        return AutogradTensor(x.child.transpose(*dims), requires_grad=x.requires_grad)

    def _backward(self, grad: ndarray, backprop_id: UUID) -> None:

        if self.x.requires_grad:

            grad = grad.transpose(*self.reverse_dims)

            self.x.add_grad(grad)

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)



# stdlib
from typing import Any
from uuid import UUID

# third party
import numpy as np

# relative
from .....core.common.serde.serializable import serializable
from ...passthrough import is_acceptable_simple_type  # type: ignore
from ..tensor import AutogradTensor
from .op import Op


@serializable(recursive_serde=True)
class AddOp(Op):
    """Sumation operation with 2 tensors"""

    __attr_allowlist__: Any = []

    def forward(self, x: AutogradTensor, y: AutogradTensor) -> AutogradTensor:  # type: ignore
        self.x = x
        self.y = y

        requires_grad = x.requires_grad

        if is_acceptable_simple_type(y):
            return AutogradTensor(x.child + y, requires_grad=requires_grad)

        requires_grad = requires_grad or y.requires_grad
        return AutogradTensor(x.child + y.child, requires_grad=requires_grad)

    def _backward(self, grad: np.ndarray, backprop_id: UUID) -> None:
        if self.x.requires_grad:
            # as we have matrix operation one of the parameters can
            # have partial shape in such scenarion we need to sum
            # gradient values by missed axis
            if self.x.shape != grad.shape:
                axis = np.argmax(np.abs(np.array(self.x.shape) - np.array(grad.shape)))
                self.x.add_grad(grad.sum(axis=axis, keepdims=True))
            else:
                self.x.add_grad(grad)
            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)

        # TODO: @Madhava do we need to use this check everywhere?
        # if isinstance(self.y, AutogradTensor) and self.y.requires_grad:
        if self.y.requires_grad:
            if self.y.shape != grad.shape:
                axis = np.argmax(np.abs(np.array(self.y.shape) - np.array(grad.shape)))
                self.y.add_grad(grad.sum(axis=axis, keepdims=True))
            else:
                self.y.add_grad(grad)
            if self.y.grad_fn:
                self.y.backward(backprop_id=backprop_id)



# stdlib
from typing import Any
from typing import Dict
from typing import List

# relative
from .passthrough import SupportedChainType  # type: ignore


class ScalarChainManagerTensor:
    """Supports convenience methods for scalar chains of abstraction"""

    def __init__(self) -> None:
        pass

    def push_abstraction_top(
        self, scalar_type: Any, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> None:
        """ """


class TensorChainManager:
    def __init__(self, child: SupportedChainType) -> None:
        self.child = child

    def push_abstraction_top(
        self, tensor_type: Any, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> None:
        """ """
        self.child = tensor_type(self.child, *args, **kwargs)

    def replace_abstraction_top(
        self, tensor_type: Any, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> None:
        self.child = tensor_type(*args, **kwargs)



# stdlib
import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type

# relative
from .....grid import GridURL
from .....lib.python.util import upcast
from ...abstract.node import AbstractNodeClient
from ..action.exception_action import ExceptionMessage
from ..node_service.generic_payload.messages import GenericPayloadMessageWithReply
from ..node_service.vpn.vpn_messages import VPNJoinMessageWithReply
from ..node_service.vpn.vpn_messages import VPNStatusMessageWithReply


class VPNAPI:
    def __init__(self, client: AbstractNodeClient):
        self.client = client

    def join_network_vpn(self, grid_url: GridURL) -> None:
        reply = self.perform_api_request(
            syft_msg=VPNJoinMessageWithReply, content={"grid_url": grid_url}
        )
        logging.info(reply.payload)
        status = "error"
        try:
            status = str(reply.payload.kwargs.get("status"))
        except Exception:  # nosec
            pass

        if status == "ok":
            print(f"🔌 {self.client} successfully connected to the VPN: {grid_url}")
        else:
            print(f"❌ {self.client} failed to connect to the VPN: {grid_url}")

    def get_status(self) -> Dict[str, Any]:
        reply = self.perform_api_request(syft_msg=VPNStatusMessageWithReply, content={})
        logging.info(reply.payload)
        try:
            return upcast(reply.payload.kwargs)
        except Exception:  # nosec
            pass

        print(f"❌ {self.client} VPN Status failed")
        return {"status": "error"}

    def perform_api_request(
        self,
        syft_msg: Optional[Type[GenericPayloadMessageWithReply]],
        content: Optional[Dict[Any, Any]] = None,
    ) -> Any:
        if syft_msg is None:
            raise ValueError(
                "Can't perform this type of api request, the message is None."
            )
        else:
            syft_msg_constructor = syft_msg

        if content is None:
            content = {}
        signed_msg = (
            syft_msg_constructor(kwargs=content)
            .to(address=self.client.address, reply_to=self.client.address)
            .sign(signing_key=self.client.signing_key)
        )  # type: ignore
        response = self.client.send_immediate_msg_with_reply(msg=signed_msg)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response



# stdlib
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from .....proto.core.node.common.action.action_sequence_pb2 import (
    ActionSequence as ActionSequence_PB,
)
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply
from .save_object_action import SaveObjectAction


@serializable()
class ActionSequence(ImmediateActionWithoutReply):
    def __init__(
        self,
        obj_lst: List[SaveObjectAction],
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.obj_lst = obj_lst

    def __repr__(self) -> str:
        obj_str = str(self.obj_lst)
        # make obj_str of reasonable length, if too long: cut into begin and end
        neg_index = max(-50, -len(obj_str) + 50)
        obj_str = obj_str = (
            obj_str[:50]
            if len(obj_str) < 50
            else obj_str[:50] + " ... " + obj_str[neg_index:]
        )
        return f"ActionSequence {obj_str}"

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        for obj in self.obj_lst:
            obj.execute_action(node=node, verify_key=verify_key)

    def _object2proto(self) -> ActionSequence_PB:
        obj_lst = list(map(lambda x: sy.serialize(x), self.obj_lst))
        addr = sy.serialize(self.address)
        return ActionSequence_PB(obj=obj_lst, address=addr)

    @staticmethod
    def _proto2object(proto: ActionSequence_PB) -> "ActionSequence":
        obj_lst = list(map(lambda x: sy.deserialize(blob=x), proto.obj))
        addr = sy.deserialize(blob=proto.address)
        return ActionSequence(obj_lst=obj_lst, address=addr)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return ActionSequence_PB



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/service/signaling_service.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.node.common import (
    metadata_pb2 as proto_dot_core_dot_node_dot_common_dot_metadata__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/grid/service/signaling_service.proto\x12\x11syft.grid.service\x1a%proto/core/common/common_object.proto\x1a%proto/core/node/common/metadata.proto\x1a\x1bproto/core/io/address.proto"\x90\x01\n\x16RegisterNewPeerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"|\n\x1aPeerSuccessfullyRegistered\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07peer_id\x18\x03 \x01(\t"\xd8\x01\n\x16SignalingAnswerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x36\n\rhost_metadata\x18\x04 \x01(\x0b\x32\x1f.syft.core.node.common.Metadata\x12\x13\n\x0btarget_peer\x18\x05 \x01(\t\x12\x11\n\thost_peer\x18\x06 \x01(\t"\xd7\x01\n\x15SignalingOfferMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x36\n\rhost_metadata\x18\x04 \x01(\x0b\x32\x1f.syft.core.node.common.Metadata\x12\x13\n\x0btarget_peer\x18\x05 \x01(\t\x12\x11\n\thost_peer\x18\x06 \x01(\t"\xb9\x01\n\x17OfferPullRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0btarget_peer\x18\x04 \x01(\t\x12\x11\n\thost_peer\x18\x05 \x01(\t"\xba\x01\n\x18\x41nswerPullRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0btarget_peer\x18\x04 \x01(\t\x12\x11\n\thost_peer\x18\x05 \x01(\t"j\n\x19SignalingRequestsNotFound\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address"g\n\x16InvalidLoopBackRequest\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address"g\n\x16\x43loseConnectionMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_REGISTERNEWPEERMESSAGE = DESCRIPTOR.message_types_by_name["RegisterNewPeerMessage"]
_PEERSUCCESSFULLYREGISTERED = DESCRIPTOR.message_types_by_name[
    "PeerSuccessfullyRegistered"
]
_SIGNALINGANSWERMESSAGE = DESCRIPTOR.message_types_by_name["SignalingAnswerMessage"]
_SIGNALINGOFFERMESSAGE = DESCRIPTOR.message_types_by_name["SignalingOfferMessage"]
_OFFERPULLREQUESTMESSAGE = DESCRIPTOR.message_types_by_name["OfferPullRequestMessage"]
_ANSWERPULLREQUESTMESSAGE = DESCRIPTOR.message_types_by_name["AnswerPullRequestMessage"]
_SIGNALINGREQUESTSNOTFOUND = DESCRIPTOR.message_types_by_name[
    "SignalingRequestsNotFound"
]
_INVALIDLOOPBACKREQUEST = DESCRIPTOR.message_types_by_name["InvalidLoopBackRequest"]
_CLOSECONNECTIONMESSAGE = DESCRIPTOR.message_types_by_name["CloseConnectionMessage"]
RegisterNewPeerMessage = _reflection.GeneratedProtocolMessageType(
    "RegisterNewPeerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _REGISTERNEWPEERMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.RegisterNewPeerMessage)
    },
)
_sym_db.RegisterMessage(RegisterNewPeerMessage)

PeerSuccessfullyRegistered = _reflection.GeneratedProtocolMessageType(
    "PeerSuccessfullyRegistered",
    (_message.Message,),
    {
        "DESCRIPTOR": _PEERSUCCESSFULLYREGISTERED,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.PeerSuccessfullyRegistered)
    },
)
_sym_db.RegisterMessage(PeerSuccessfullyRegistered)

SignalingAnswerMessage = _reflection.GeneratedProtocolMessageType(
    "SignalingAnswerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNALINGANSWERMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.SignalingAnswerMessage)
    },
)
_sym_db.RegisterMessage(SignalingAnswerMessage)

SignalingOfferMessage = _reflection.GeneratedProtocolMessageType(
    "SignalingOfferMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNALINGOFFERMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.SignalingOfferMessage)
    },
)
_sym_db.RegisterMessage(SignalingOfferMessage)

OfferPullRequestMessage = _reflection.GeneratedProtocolMessageType(
    "OfferPullRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _OFFERPULLREQUESTMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.OfferPullRequestMessage)
    },
)
_sym_db.RegisterMessage(OfferPullRequestMessage)

AnswerPullRequestMessage = _reflection.GeneratedProtocolMessageType(
    "AnswerPullRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _ANSWERPULLREQUESTMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.AnswerPullRequestMessage)
    },
)
_sym_db.RegisterMessage(AnswerPullRequestMessage)

SignalingRequestsNotFound = _reflection.GeneratedProtocolMessageType(
    "SignalingRequestsNotFound",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNALINGREQUESTSNOTFOUND,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.SignalingRequestsNotFound)
    },
)
_sym_db.RegisterMessage(SignalingRequestsNotFound)

InvalidLoopBackRequest = _reflection.GeneratedProtocolMessageType(
    "InvalidLoopBackRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _INVALIDLOOPBACKREQUEST,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.InvalidLoopBackRequest)
    },
)
_sym_db.RegisterMessage(InvalidLoopBackRequest)

CloseConnectionMessage = _reflection.GeneratedProtocolMessageType(
    "CloseConnectionMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CLOSECONNECTIONMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.CloseConnectionMessage)
    },
)
_sym_db.RegisterMessage(CloseConnectionMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _REGISTERNEWPEERMESSAGE._serialized_start = 173
    _REGISTERNEWPEERMESSAGE._serialized_end = 317
    _PEERSUCCESSFULLYREGISTERED._serialized_start = 319
    _PEERSUCCESSFULLYREGISTERED._serialized_end = 443
    _SIGNALINGANSWERMESSAGE._serialized_start = 446
    _SIGNALINGANSWERMESSAGE._serialized_end = 662
    _SIGNALINGOFFERMESSAGE._serialized_start = 665
    _SIGNALINGOFFERMESSAGE._serialized_end = 880
    _OFFERPULLREQUESTMESSAGE._serialized_start = 883
    _OFFERPULLREQUESTMESSAGE._serialized_end = 1068
    _ANSWERPULLREQUESTMESSAGE._serialized_start = 1071
    _ANSWERPULLREQUESTMESSAGE._serialized_end = 1257
    _SIGNALINGREQUESTSNOTFOUND._serialized_start = 1259
    _SIGNALINGREQUESTSNOTFOUND._serialized_end = 1365
    _INVALIDLOOPBACKREQUEST._serialized_start = 1367
    _INVALIDLOOPBACKREQUEST._serialized_end = 1470
    _CLOSECONNECTIONMESSAGE._serialized_start = 1472
    _CLOSECONNECTIONMESSAGE._serialized_end = 1575
# @@protoc_insertion_point(module_scope)



# stdlib
import json
from typing import Dict
from typing import List as TypeList
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.grid.messages.setup_messages_pb2 import (
    CreateInitialSetUpMessage as CreateInitialSetUpMessage_PB,
)
from ......proto.grid.messages.setup_messages_pb2 import (
    GetSetUpMessage as GetSetUpMessage_PB,
)
from ......proto.grid.messages.setup_messages_pb2 import (
    GetSetUpResponse as GetSetUpResponse_PB,
)
from ......proto.grid.messages.setup_messages_pb2 import (
    UpdateSetupMessage as UpdateSetupMessage_PB,
)
from ......proto.grid.messages.setup_messages_pb2 import (
    UpdateSetupResponse as UpdateSetupResponse_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
@final
class GetSetUpMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetSetUpMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetSetUpMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetSetUpMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetSetUpMessage_PB,
    ) -> "GetSetUpMessage":
        """Creates a GetSetUpMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetSetUpMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetSetUpMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetSetUpMessage_PB


@serializable()
@final
class GetSetUpResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> GetSetUpResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetSetUpResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetSetUpResponse_PB,
    ) -> "GetSetUpResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetSetUpResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetSetUpResponse_PB


@serializable()
@final
class CreateInitialSetUpMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        name: str,
        email: str,
        password: str,
        domain_name: str,
        budget: float,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.name = name
        self.email = email
        self.password = password
        self.domain_name = domain_name
        self.budget = budget

    def _object2proto(self) -> CreateInitialSetUpMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateInitialSetUpMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateInitialSetUpMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            name=self.name,
            email=self.email,
            password=self.password,
            domain_name=self.domain_name,
            budget=self.budget,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: CreateInitialSetUpMessage_PB,
    ) -> "CreateInitialSetUpMessage":
        """Creates a CreateInitialSetUpMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateInitialSetUpMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateInitialSetUpMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            name=proto.name,
            email=proto.email,
            password=proto.password,
            budget=proto.budget,
            domain_name=proto.domain_name,
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateInitialSetUpMessage_PB


@serializable()
@final
class UpdateSetupMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        domain_name: str,
        description: str,
        daa: bool,
        contact: str,
        reply_to: Address,
        daa_document: Optional[bytes] = b"",
        tags: Optional[TypeList] = None,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.daa = daa
        self.contact = contact
        self.description = description
        self.domain_name = domain_name
        self.daa_document = daa_document
        self.tags = tags if tags is not None else []

    def _object2proto(self) -> UpdateSetupMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: UpdateSetupMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateSetupMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            domain_name=self.domain_name,
            contact=self.contact,
            daa=self.daa,
            description=self.description,
            daa_document=self.daa_document,
            tags=self.tags,
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateSetupMessage_PB,
    ) -> "UpdateSetupMessage":
        """Creates a UpdateSetupMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: UpdateSetupMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateSetupMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            daa=proto.daa,
            contact=proto.contact,
            domain_name=proto.domain_name,
            description=proto.description,
            daa_document=proto.daa_document,
            tags=[tag for tag in proto.tags],
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateSetupMessage_PB


@serializable()
@final
class UpdateSetupResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> UpdateSetupResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateSetupResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateSetupResponse_PB,
    ) -> "UpdateSetupResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateSetupResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateSetupResponse_PB
